// $Id: AlignAlgorithm.h 1151 2021-01-29 00:45:32Z ge $
/// \file AlignAlgorithm.h
/// \brief contains the AlignAlgorithm class
///
/// $Revision: 1151 $
/// \author Gerald Weber <gweberbh@gmail.com>

#ifndef ALIGNALGORITHM_H
#define ALIGNALGORITHM_H "$Id: AlignAlgorithm.h 1151 2021-01-29 00:45:32Z ge $"

#include "NucleotideSequence.h"
#include "BasePair.h"
#include "valmatrix.h"
#include "stdmatrix.h"
#include <vector>
#include <deque>
#include <utility>

namespace gbc_exp {
using namespace gbc;

template<class _Array_Tp=valmatrix<int>, class _Align_Array_Tp=_Array_Tp >
/// \brief A class that aligns sequences using the Needleman Wunsch algorithm.
/// To change the score_type change the array_type, for example valmatrix<double>
/// will result in score_type being double.
class NeedlemanWunsch
  {
  public:
    typedef int                sequence_type;    ///< Type of sequence
    typedef _Array_Tp          array_type;
    typedef _Align_Array_Tp    align_array_type;
    typedef typename align_array_type::value_type score_type; ///< Type of alignment score, results from align_array_type
    typedef std::pair<size_t,size_t> position_type;           ///< Type of position in matrix
    
    enum DisplayAlignmentType { MainTop=0, TargetTop =1 }; 
    
    DisplayAlignmentType DisplayAlignment;
    
    align_array_type AlignmentMatrix;            ///< Main alignment matrix
    array_type MainMatrix;                       ///< Main sequence in number format
    array_type TargetMatrix;                     ///< Target sequence in number format
    score_type match;                            ///< The score of a match of two nucleotides
    score_type mismatch;                         ///< The score of a mismatch of two nucleotides
    score_type gap;                              ///< The score of a gap in the alignment
    score_type initialize;
    size_t MainSize;
    size_t TargetSize;
    size_t SmallestSize;
    double Score_limit;                          ///< This controls the minimal score to continue filling the matrix, used with Early_break
    bool Early_break;                            ///< Controls if an early break in the matrix filling process is to be used.
    bool Record_backtrace;
    score_type diagonal, top, left;
    int OutputContigMargin;
    int RecurrenceDirection;
    std::string TopComment, BottomComment;       ///> Additional comments when printing alignment
    std::string Top_string, Middle_string, Bottom_string;
    int Mismatch_count, LargestConsecutiveMismatch;
    long int Sequence_position;
    NucleotideSequence<sequence_type> MainSequence;
    NucleotideSequence<sequence_type> Target;
    position_type pos;
    std::deque<size_t> top_alignment, bottom_alignment; ///< The alignment queues sequence positions
    std::deque<score_type>  scores;
    bool target_backtrace_incomplete; ///< used by SmithWatermanBP

  NeedlemanWunsch(void)
    :DisplayAlignment(), match(2), mismatch(-1), gap(-2), initialize(0), MainSize(0), TargetSize(0), 
     Score_limit(0.0), Early_break(false), Record_backtrace(true),
     diagonal(0), top(0), left(0), OutputContigMargin(0), RecurrenceDirection(0), TopComment(), BottomComment(),
     Top_string(), Middle_string(), Bottom_string(), Mismatch_count(0), LargestConsecutiveMismatch(0), target_backtrace_incomplete(true)
     {
     MainSequence.Extended_strand_interpretation = false;
     Target.Extended_strand_interpretation = false;
    }

  virtual ~NeedlemanWunsch(void) {}

  inline virtual void reset(void) {}
  
  inline void set_display_main_top(void) {DisplayAlignment=MainTop;}

  /// \brief Prepares the matrices for the alignment
  ///
  /// This function allocates sufficient memory for the main sequence and the alignment
  /// matrix. It also assigns the nucleotide sequence to the numeric array.
  inline virtual void prepare_main_sequence(void) 
    {
    if (MainSize != MainSequence.size()) 
      {
      MainSize=MainSequence.size();
      MainMatrix.resize(1,MainSize+1);
      AlignmentMatrix.resize(TargetSize+1,MainSize+1);
      if (highest_possible_score() > std::numeric_limits<score_type>::max())
        {
        std::cerr << "The selected score may exceed limits" << std::endl;
        }
      }
    for (size_t i=1; i <= MainSize; ++i) MainMatrix[i]=MainSequence[i-1].representation;
    }

  /// \brief Prepares the matrices for the alignment
  ///
  /// This function has a similar role as prepare_main_sequence().
  inline virtual void prepare_target_sequence(void) 
    {
    if (TargetSize != Target.size()) 
      {
      TargetSize=Target.size();
      TargetMatrix.resize(1,TargetSize+1);
      AlignmentMatrix.resize(TargetSize+1,MainSize+1);
      if (highest_possible_score() > std::numeric_limits<score_type>::max())
        {
        std::cerr << "The selected score may exceed limits" << std::endl;
        }
      }
    for (size_t i=1; i <= TargetSize; ++i) TargetMatrix[i]=Target[i-1].representation;
    }

  inline virtual score_type rule(const size_t &i, const size_t &j) const
    {
    if (TargetMatrix[i] == MainMatrix[j]) return match;
    else                                  return mismatch;
    }

  /// \brief Provides the overall alingment score
  ///
  /// For the Needleman-Wunsch algorithm this is the last position of the matrix.
  inline virtual score_type alignment_score(void) {return AlignmentMatrix(TargetSize,MainSize);}

  /// \brief Returns the highest possible score
  ///
  /// For a given scheme of match, mismatch and gap determine the 
  /// highest score an alignment could possibly have. 
  /// Considering that a perfect alignment would occur along a diagonal we
  /// divide by the smallest of TargetSize and MainSize
  inline virtual score_type highest_possible_score(void) 
    {
    return std::max(match,std::max(mismatch,gap))*std::min(TargetSize,MainSize);
    }

  /// \brief Returns the lowest possible score
  inline virtual score_type lowest_possible_score(void) 
    {
    return std::min(match,std::min(mismatch,gap))*std::max(TargetSize,MainSize);
    }

  inline virtual void score_limit(double f) {Score_limit=f;}

  /// \brief Calculates the fraction of score and maximum possible score
  /// \returns A value smaller or equal 1.0
  inline virtual double relative_score(void) 
   {
   return static_cast<double>(alignment_score())/(static_cast<double>(highest_possible_score()));
   }

  /// \brief Determines where to start backtracing
  ///
  /// For the Needleman-Wunsch algorithm, backtracing starts at the
  /// last position of the matrix. This may be different for other 
  /// algorithms, therefore this function is virtual.
  inline virtual position_type backtrace_start_position(void) 
    {return position_type(TargetSize,MainSize);}

  /// \brief Checks if backtracing can go further
  inline virtual bool backtrace_not_finished(position_type& pos)
    {
    return ((pos.first > 0) && (pos.second > 0));
    }

  /// \brief The main backtrace algorithm
  inline virtual void backtrace(void)
    {
    top_alignment.clear();
    bottom_alignment.clear();
    scores.clear();

    std::pair<sequence_type,sequence_type> p(TargetSize,MainSize);
    position_type from, from_top, from_left, from_diagonal;
    pos=backtrace_start_position();
    while (backtrace_not_finished(pos))
      {
      // first we record the positions 
      from_top=from_left=from_diagonal=pos;
      from_top.second--;
      from_left.first--;
      from_diagonal.first--;
      from_diagonal.second--;

      // now we rerun the recurrence with the current position
      // this calculates the values top, left and right which are
      // then compared to rec.
      recurrence(pos.first,pos.second);
      int rec=RecurrenceDirection;//

      if (rec==0)
        {
        from=from_diagonal;
        p.first=Target[pos.first-1]; p.second=MainSequence[pos.second-1];
        }
      else
        {
        if (rec==-1)
          {
          from=from_top;
          p.first=0; // No match 
          p.second=MainSequence[pos.second-1];
          }
        else
          {
          if (rec==1)
            {
            from=from_left;
            p.first=Target[pos.first-1]; 
            p.second=0; // No match
            }
          else std::cout << "(Error) wrong RecurrenceDirection, requires program debug" << rec << std::endl;
          }
        }
      // the new position is wherever we've came from. Note that this implementation
      // privileges coming from left. If backtrace has finished this value is used to 
      // determine the start of the backtrace.
      pos=from; 

      // Recording the backtrace may be potentially memory consuming for very large alignments
      if (Record_backtrace)
        {
        bottom_alignment.push_front(pos.second); 
        top_alignment.push_front(pos.first);
//         std::cout << pos.first << " " << p.first << " " << pos.second << " " << p.second << std::endl;
        scores.push_front(rec);
        }
      }
    }

  /// \brief Returns the alignment start position
  ///
  /// This function only returns a meaningful value after running backtrace.
  int start_position(void) {return pos.second;}

  template<class _StreamTp>
  void print_alignment(_StreamTp &out)
    {
    out << start_position() << " (" << start_position()+1 << ")" << std::endl;
    Top_string=Middle_string=Bottom_string=std::string();
    Mismatch_count=LargestConsecutiveMismatch=0;
    for(auto it : top_alignment)
      {
      Top_string += Nucleotide<sequence_type>::dictionary.invert(Target[it]);
      }
    out << std::endl; 
    int matches=0, consecutive_mismatches=0;
    std::string show_match;
    for(auto it=top_alignment.begin(), it2=bottom_alignment.begin(); it != top_alignment.end() and it2 != bottom_alignment.end(); ++it, ++it2)
      {
      if (rule(*it+1,*it2+1) == match) 
        {
        show_match+= "|"; 
        matches++;
        if (consecutive_mismatches > 0)
          {
          if (consecutive_mismatches > LargestConsecutiveMismatch) LargestConsecutiveMismatch=consecutive_mismatches;
          consecutive_mismatches=0;
          }
        }
      else 
        {
        show_match+= "X"; 
        Mismatch_count++; 
        consecutive_mismatches++;
        }
      }
    if (consecutive_mismatches > LargestConsecutiveMismatch) LargestConsecutiveMismatch=consecutive_mismatches;
    if (matches > 0) Middle_string += show_match;
    for(auto it : bottom_alignment)
      {
      Bottom_string += Nucleotide<sequence_type>::dictionary.invert(MainSequence[it]);
      }
    if (DisplayAlignment == TargetTop) out << Top_string    << TopComment << std::endl << Middle_string << std::endl << Bottom_string << BottomComment << std::endl;
    else                               out << Bottom_string << TopComment << std::endl << Middle_string << std::endl << Top_string    << BottomComment << std::endl;
    }
    
  inline virtual void print_matrix(std::ostream &out)
    {
    out << "  ";
    size_t i,j;
    for (j=0; j < MainSize; j++)
      out << MainSequence[j] << " ";
    out << std::endl;
    for (i=0; i < TargetSize; i++)
      {
      out << Target[i] << " ";
      for (j=0; j < MainSize; j++)
        out << AlignmentMatrix(i+1,j+1) << " ";
      out << std::endl;
      }
    }

  /// \brief The recurrence function is the heart of the algorithm
  ///
  /// This function also sets the variables, diagonal, left and top
  /// which are necessary for the bactrace algorithm.
  inline virtual score_type recurrence(const size_t &i, const size_t &j)
    {
    diagonal=AlignmentMatrix(i-1,j-1)+rule(i,j);
    left    =AlignmentMatrix(i-1,j)  +gap;
    top     =AlignmentMatrix(i,j-1)  +gap;
    score_type maximum=diagonal; RecurrenceDirection=0;
    if (left > maximum) {maximum=left; RecurrenceDirection=1;}
    if (top  > maximum) {maximum=top;  RecurrenceDirection=-1;}
    return maximum;
    }

  /// \brief This method iterates over the matrix positions applying the recurrence relation
  ///
  /// To shorten the execution time it is possible to set a Score_limit which
  /// stops the matrix filling process if a minimal score is not obtainable.
  /// If Early_break is set to false Score_limit is ignored.
  inline virtual void fill_matrix(void)
    {
    reset();

    score_type limit=static_cast<score_type>(round(TargetSize*match*Score_limit));

    AlignmentMatrix(0,0)=initialize;
    for (size_t i=1; i <= TargetSize; ++i)
      {
      AlignmentMatrix(i,0)=initialize;
      for (size_t j=1; j <= MainSize; ++j)
        {
        AlignmentMatrix(i,j)=recurrence(i,j);
        }
      if (Early_break) 
        {
        //If minimum score is impossible to achieve, give up
        if ((static_cast<score_type>(TargetSize-i)*match+alignment_score()) < limit ) break; 
        }
      }
    }
  };

template<class _Array_Tp=valmatrix<int> >
/// \brief This class implements a NeedlemanWunsh algorithm with little memory usage.
///
/// We use only a 2xN matrix, this class has no backtrace capability.
class ThinNeedlemanWunsch: public NeedlemanWunsch<_Array_Tp>
  {
  public:
    typedef NeedlemanWunsch<_Array_Tp> base_type; 
    typedef typename base_type::score_type    score_type;
    typedef typename base_type::position_type position_type;
    using base_type::AlignmentMatrix;
    using base_type::diagonal;
    using base_type::Early_break;
    using base_type::gap;
    using base_type::highest_possible_score;
    using base_type::initialize;
    using base_type::left;
    using base_type::MainMatrix;
    using base_type::MainSize;
    using base_type::MainSequence;
    using base_type::match;
    using base_type::reset;
    using base_type::rule;
    using base_type::Score_limit;
    using base_type::TargetSize;
    using base_type::Target;
    using base_type::TargetMatrix;
    using base_type::top;

  ThinNeedlemanWunsch(void) {}

  virtual ~ThinNeedlemanWunsch(void) {}

  /// \brief Prepares the matrices for the alignment
  ///
  /// This function allocates sufficient memory for the main sequence and the alignment
  /// matrix. It also assigns the nucleotide sequence to the numeric array.
  inline virtual void prepare_main_sequence(void) 
    {
    if (MainSize != MainSequence.size()) 
      {
      MainSize=MainSequence.size();
      MainMatrix.resize(1,MainSize+1);
      AlignmentMatrix.resize(2,MainSize+1);
      if (highest_possible_score() > std::numeric_limits<score_type>::max())
        {
        std::cerr << "The selected score may exceed limits" << std::endl;
        }
      }
    for (size_t i=1; i <= MainSize; ++i) MainMatrix[i]=MainSequence[i-1].representation;
    }

  /// \brief Prepares the matrices for the alignment
  ///
  /// This function has a similar role as prepare_main_sequence() but does not resize AlignmentMatrix
  inline virtual void prepare_target_sequence(void) 
    {
    if (TargetSize != Target.size()) 
      {
      TargetSize=Target.size();
      TargetMatrix.resize(1,TargetSize+1);
      AlignmentMatrix.resize(2,MainSize+1);
      if (highest_possible_score() > std::numeric_limits<score_type>::max())
        {
        std::cerr << "The selected score may exceed limits" << std::endl;
        }
      }
    for (size_t i=1; i <= TargetSize; ++i) TargetMatrix[i]=Target[i-1].representation;
    }

  inline size_t odd(const size_t& i)
    {return (i/2)*2!=i;}

  inline size_t ii(const size_t& i)
    {
    if (odd(i)) return 1;
      else      return 0;
    }

  inline size_t i_minus_1(const size_t& i)
    {
    if (odd(i)) return 0;
      else      return 1;
    }


  /// \brief The recurrence function is the heart of the algorithm
  ///
  /// This function also sets the variables, diagonal, left and top
  /// which are necessary for the bactrace algorithm.
  inline virtual score_type recurrence(const size_t &i, const size_t &j)
    {
    
    diagonal=AlignmentMatrix(0,j-1)+rule(i,j);
    left    =AlignmentMatrix(0,j)  +gap;
    top     =AlignmentMatrix(1,j-1)  +gap;
    score_type maximum=diagonal;
    if (left > maximum) maximum=left;
    if (top  > maximum) maximum=top;
    return maximum;
    }

  /// \brief This method iterates over the matrix positions applying the recurrence relation
  ///
  /// To shorten the execution time it is possible to set a Score_limit which
  /// stops the matrix filling process if a minimal score is not obtainable.
  /// If Early_break is set to false Score_limit is ignored.
  inline virtual void fill_matrix(void)
    {
    reset();

    score_type limit=static_cast<score_type>(round(TargetSize*match*Score_limit));

    for (size_t j=0; j <= MainSize; ++j)
      {AlignmentMatrix(0,j)=initialize;}
    for (size_t i=1; i <= TargetSize; ++i)
      {
      AlignmentMatrix(1,0)=initialize;
      for (size_t j=1; j <= MainSize; ++j)
        {
        AlignmentMatrix(1,j)=recurrence(i,j);
        }
      for (size_t j=1; j <= MainSize; ++j)
        {
        AlignmentMatrix(0,j)=AlignmentMatrix(1,j);
        }
      if (Early_break) 
        {
        //If minimum score is impossible to achieve, give up
        if ((static_cast<score_type>(TargetSize-i)*match+alignment_score()) < limit ) break; 
        }
      }
    }

  /// \brief Provides the overall alingment score
  ///
  /// For the Needleman-Wunsch algorithm this is the last position of the matrix.
  inline virtual score_type alignment_score(void) 
    {return AlignmentMatrix(1,MainSize);}

  };


template<class _Array_Tp=valmatrix<int>, class _Score_Tp=int >
/// \brief A class that aligns sequences using the Smith Waterman algorithm.
///
class SmithWaterman: virtual public NeedlemanWunsch<_Array_Tp>
  {
  public:
    typedef NeedlemanWunsch<_Array_Tp> base_type; 
    using base_type::diagonal;
    using base_type::left;
    using base_type::top;
    using base_type::gap;
    using base_type::AlignmentMatrix;
    using base_type::rule;

    typedef typename base_type::score_type    score_type;
    typedef typename base_type::position_type position_type;

    score_type    Highest_alignment;
    position_type Highest_pos;

  SmithWaterman(void)
    :Highest_alignment(0), Highest_pos(0,0) {}

  virtual ~SmithWaterman(void) {}

  inline virtual void reset(void) {Highest_alignment=0;}

  inline virtual score_type recurrence(const size_t &i, const size_t &j)
    {
    diagonal=AlignmentMatrix(i-1,j-1)+rule(i,j);
    left    =AlignmentMatrix(i-1,j)  +gap;
    top     =AlignmentMatrix(i,j-1)  +gap;
    score_type maximum=score_type();
    if (diagonal > maximum) maximum=diagonal;
    if (left     > maximum) maximum=left;
    if (top      > maximum) maximum=top;
    if (maximum > Highest_alignment) 
      {
      Highest_alignment=maximum;
      Highest_pos=position_type(i,j);
      }
    return maximum;
    }

  inline virtual score_type alignment_score(void) {return AlignmentMatrix(Highest_pos.first,Highest_pos.second);}

  inline virtual position_type backtrace_start_position(void) 
    {return Highest_pos;}

  inline virtual bool backtrace_not_finished(position_type& pos)
    {
    return (AlignmentMatrix(pos.first,pos.second) > 0) &&
           (base_type::backtrace_not_finished(pos));
    }


  };

template<class _Array_Tp=valmatrix<int>, class _Score_Tp=int >
/// \brief A class that aligns sequences to matching base pairs using the Smith Waterman algorithm.
///
class SmithWatermanBP: virtual public SmithWaterman<_Array_Tp>
  {
  public:
    typedef SmithWaterman<_Array_Tp,_Score_Tp> base_type; 
    using base_type::MainMatrix;
    using base_type::MainSequence;
    using base_type::TargetMatrix;
    using base_type::Target;
    using base_type::TargetSize;
    using base_type::match;
    using base_type::mismatch;
    using base_type::Highest_pos;
    using base_type::target_backtrace_incomplete;

    typedef typename base_type::score_type    score_type;
    typedef typename base_type::position_type position_type;
    typedef std::deque<int> bp_deque_type;
   
  bp_deque_type acceptable_bp;

  SmithWatermanBP(void)
    {
    BasePair<int> BP("AT",BasePair<int>::simplify_symmetry);
    BP.reduce_to_smallest_symmetry();
    acceptable_bp.push_back(BP.first.representation + BP.second.representation);
    
    BP.string_interpretation("CG");
    BP.reduce_to_smallest_symmetry();
    acceptable_bp.push_back(BP.first.representation + BP.second.representation);
    }

  virtual ~SmithWatermanBP(void) {}

  inline virtual score_type rule(const size_t &i, const size_t &j) const
    {
    int rep=Target[i-1].representation+MainSequence[j-1].representation;
    bool found=false;
    for (auto bpl = acceptable_bp.begin(); bpl != acceptable_bp.end(); bpl++)
      {
      if (rep == *bpl) {found=true; break; }
      }
        
    if (found) return match;
    else return mismatch;
    }
    
  inline virtual position_type backtrace_start_position(void) 
    {return position_type(TargetSize,Highest_pos.second);}

// incomplete target backtrace can happen when a partial alignment happens at the beginning of the genome
  inline virtual bool backtrace_not_finished(position_type& pos)
    {
    target_backtrace_incomplete = (pos.first == 0) and (pos.second > 0); //target (pos.second) could not be fully tracebacked
    return ((pos.first > 0) && (pos.second > 0));
    }

 
  };
};
#endif
