// $Id: AlignSequence.h 1120 2020-05-06 00:10:57Z ge $
/// \file AlignSequence.h
/// \brief contains the AlignSequence class
///
/// $Revision: 1120 $
/// \author Gerald Weber <gweberbh@gmail.com>

#ifndef ALIGNSEQUENCE_H
#define ALIGNSEQUENCE_H "$Id: AlignSequence.h 1120 2020-05-06 00:10:57Z ge $"

#include "NucleotideSequence.h"
#include "valmatrix.h"
#include "stdmatrix.h"
#include <vector>
#include <deque>
#include <utility>
#include "AlignAlgorithm.h"
#include "RegexAux.h"

namespace gbc_exp {
using namespace gbc;

template<class _AlignmentType = NeedlemanWunsch<> >
/// \brief A class that aligns sequences.
///
class AlignSequence: public  _AlignmentType
  {
  public:
    typedef typename _AlignmentType::sequence_type sequence_type;    ///< Type of sequence.
    typedef typename _AlignmentType::score_type    score_type;
    typedef _AlignmentType                alignment_type;
    std::ostream *ReportFile;             ///< This file usually holds information for debuging or controlling
    std::ostream *FastaFile;              ///< Outputs the found sequence
    std::ostream *BacktraceFile;          ///< Prints backtracing information, this can be very large
    std::ostream *TfRegFile;              ///< Prints sequences in a format readable by TfReg
    std::string Target_description;       ///< This description is usually the fasta file comment
    std::string MainSequence_description; ///< This description is usually the fasta file comment
    int Target_number;                    ///< Target sequence number in file
    int MainSequence_number;              ///< Main sequence number in file
    std::string Target_ID;                ///< short ID when full description is inconvenient
    std::string MainSequence_ID;          ///< short ID when full description is inconvenient
    std::string Strand_type;              ///< main, complementary, reverse, complementary-reverse
    std::string Best_strand;              ///< Strand where the best score was found
    score_type Best_score;                ///< Best score, usually in a context where several scores were calculated
    double Best_relative_score;           ///< Best relative score, usually in a context where several scores were calculated
    double Maximal_size_difference;       ///< sizes of Target and MainSequence should be no more different that this value
    std::vector<double> Relative_scores;  ///< Collection of calculated scores
    std::vector<std::string> TfRegLines;
    std::string Fasta_comment_type;       ///< Flag which tells us about fasta comment formats
    double Minimal_score_backtrace;       ///< Minimal score for which we do backtracing
    bool Minimal_report;                  ///< Prints only minimal result info for saving disk space
    bool Always_backtrace;
    std::stringstream Backtrace_stream;   ///< Stream to print backtrace into if necessary
    std::string Directions;               ///< Specifies in which directions we compare

    using alignment_type::alignment_score;
    using alignment_type::backtrace;
    using alignment_type::fill_matrix;
    using alignment_type::MainSequence;
    using alignment_type::MainSize;
    using alignment_type::OutputContigMargin;
    using alignment_type::prepare_main_sequence;
    using alignment_type::prepare_target_sequence;
    using alignment_type::relative_score;
    using alignment_type::Record_backtrace;
    using alignment_type::Score_limit;
    using alignment_type::Sequence_position;
    using alignment_type::start_position;
    using alignment_type::Target;
    using alignment_type::TargetSize;
    using alignment_type::target_backtrace_incomplete;
   
  /// The void constructor.
  AlignSequence(void): Best_score(score_type()), Best_relative_score(0.0), Maximal_size_difference(1.0),
    Relative_scores(4), Fasta_comment_type(), Minimal_score_backtrace(), Minimal_report(false), Always_backtrace(false),
    Directions("mrcd") //mrcd=main, reverse, complementary, reverse complementary (d)
    {
    ReportFile=FastaFile=BacktraceFile=TfRegFile=NULL;
    }

  inline void genome(NucleotideSequence<sequence_type>& m)
    {
    MainSequence=m;
    MainSize=MainSequence.size();
    }

  inline void target(NucleotideSequence<sequence_type>& t)
    {
    Target=t;
    TargetSize=Target.size();
    }

  void print_report(void)
    {
    *ReportFile << "Found target: " << Target_description << std::endl;
    *ReportFile << MainSequence_description << std::endl;
    *ReportFile << "Strand: " << Strand_type << std::endl;
    *ReportFile << "Score: " << alignment_score() << std::endl;
    *ReportFile << "Relative score: " << relative_score() << std::endl;

    backtrace();
    if (Record_backtrace)
      {
      *BacktraceFile << "Position: " << Sequence_position+start_position() << std::endl;
      *BacktraceFile << "Target: " << Target << std::endl;
      alignment_type::print_alignment(*BacktraceFile);
      }
    *ReportFile << "Position: " << Sequence_position+start_position() << std::endl;
    }

  /// \brief Searches for a field= otherwise return all description
  std::string extract_field(std::string field,std::string& desc)
    {
    std::string::size_type loc_p=desc.find(field+"=");
    std::string loc;
    if (loc_p != std::string::npos) 
      {
      loc=desc.substr(loc_p);
      loc=loc.substr(loc.find('=')+1);
      return loc.substr(0,loc.find(';'));
      }
    else return desc;
    }

   std::string print_tfreg_dat(std::string cm)
    {
    std::stringstream out;
    out << alignment_type::Bottom_string << " " << alignment_type::Top_string 
        << " 60.0 50 0.5 #" 
        << Target_description.substr(0,Target_description.find(' ')) << "@"
        << MainSequence_description.substr(0,MainSequence_description.find(' ')) << "@"
        << alignment_type::start_position() 
        << " mm=" << alignment_type::Mismatch_count
        << " lcm=" << alignment_type::LargestConsecutiveMismatch
        << " algn=" << alignment_type::Middle_string
        << " " << cm;
    return out.str();
    }

    
  /// \brief Reports after multi_strand_alignment
  void print_multi_strand_report(void)
    {
    if (ReportFile!=NULL)
      {
      //Accesion info
      *ReportFile << Target_description.substr(0,Target_description.find(' ')) << " "
                  << MainSequence_description.substr(0,MainSequence_description.find(' ')) << " ";
      if (Fasta_comment_type == "dmeltransposon")
         {
        //Name info
        *ReportFile << extract_field("name",Target_description) << " "
                    << extract_field("name",MainSequence_description) << " ";

        //Location info
        *ReportFile << extract_field("loc",Target_description) << " "
                    << extract_field("loc",MainSequence_description) << " ";
        }

      //Results
      *ReportFile << std::setw(5) << Target.size() << " "
                  << std::setw(5) << MainSequence.size() << " "
                  << std::setw(2) << Best_strand << " " 
                  << std::showpoint << std::setprecision(6) << Best_relative_score << " "
                  << Best_score << " "
                  << Relative_scores[0] << " "
                  << Relative_scores[1] << " "
                  << Relative_scores[2] << " "
                  << Relative_scores[3] << " "
                  << std::endl;
//        if (Minimal_score_backtrace > 0.0) 
//          {
         *ReportFile << Backtrace_stream.str() << std::endl;
         size_t i;
         for(i=0; i < TfRegLines.size(); i++)
           {
           if (target_backtrace_incomplete) *TfRegFile << TfRegLines[i] << std::endl;
           else *ReportFile << "incomplete traceback not written to TfRegFile file" << std::endl;
           }
         TfRegLines.clear();
//          }

      }
    else
      {
      std::cerr << "Report file not open" << std::endl;
      exit(1);
      }
    }

  /// \brief Reports after multi_strand_alignment
  void print_minimal_report(void)
    {
    if (ReportFile!=NULL)
      {
      *ReportFile << Target_number << " "
                  << MainSequence_number << " "
                  << std::showpoint << std::setprecision(4) << Best_relative_score 
                  << std::endl;
      }
    else
      {
      std::cerr << "Report file not open" << std::endl;
      exit(1);
      }
    }


  void print_fasta_section(void)
    {
    long int start=start_position()-OutputContigMargin;
    if (start < 0) start=0;
    long int stop=start_position()+Target.size()+OutputContigMargin;
    if (static_cast<size_t>(stop) > MainSequence.size())
      stop=MainSequence.size();
    long int p,line;
    *FastaFile << "> " 
        << relative_score() << " "
        << Sequence_position+start << ":"
        << Sequence_position+stop << " "
        << Sequence_position+start_position() << ":"
        << Sequence_position+start_position()+Target.size() << " "
        << Strand_type << " "
        << Target_description << " "
        << std::endl;
    for(p=start,line=0; p < stop; p++,line++)
      {
      *FastaFile << alignment_type::MainSequence[p];
      if (line > 80) 
        {
        *FastaFile << std::endl;
        line=-1;
        }
      }
    *FastaFile << std::endl;
    }

  void run_target_alignment(void)
    {
    alignment_type::prepare_target_sequence();
    alignment_type::fill_matrix();
    }

  void run_complete_alignment(void)
    {
    prepare_main_sequence();
    if (MainSequence.size() >= Target.size())
      {
      Strand_type="Main";
      run_target_alignment();
      Relative_scores[0]=relative_score();
      Best_relative_score=Relative_scores[0];
    
      Target.reverse();
      Strand_type="Main (reverse)";
      run_target_alignment();
      Relative_scores[1]=relative_score();
      if (Relative_scores[1] > Best_relative_score) Best_relative_score=Relative_scores[1];

      Target.complementary();
      Strand_type="Complementary (reverse)";
      run_target_alignment();
      Relative_scores[3]=relative_score();
      if (Relative_scores[3] > Best_relative_score) Best_relative_score=Relative_scores[3];

      Target.reverse();
      Strand_type="Complementary";
      run_target_alignment();
      Relative_scores[2]=relative_score();
      if (Relative_scores[2] > Best_relative_score) Best_relative_score=Relative_scores[2];

      //Changes back to original
      Target.complementary();
      Strand_type="Main";
      if (Minimal_report) print_minimal_report();
      else print_multi_strand_report();

      }
    else 
      {
      MainSequence.clear();
      }
    }


  /// \brief Runs the alignment, but prints no reports
  void run_strand_alignment(NucleotideSequence<typename alignment_type::sequence_type> *changesequence, std::string Strand)
    {
    Strand_type=Strand;
    if (changesequence == &Target) prepare_target_sequence();
    else                           prepare_main_sequence();
    fill_matrix();
    if (Best_relative_score < relative_score()) 
      {
      Best_relative_score=relative_score();
      Best_score=alignment_score();
      Best_strand=Strand;
      Score_limit=relative_score();
      if (Best_relative_score > Minimal_score_backtrace and !Always_backtrace)
        {
        backtrace();
        Backtrace_stream.str(std::string());
        alignment_type::print_alignment(Backtrace_stream);
        }
      }
    if (Always_backtrace)
      {
      backtrace();
      Backtrace_stream << Strand << std::endl;
      alignment_type::print_alignment(Backtrace_stream);
      Backtrace_stream << std::endl;
      }
      
    }

  /// \brief Tries to find best alignment over all strand directions
  void run_multi_strand_alignment(void)
    {
    Best_relative_score=-std::numeric_limits<typename alignment_type::score_type>::max();
    Best_strand="X ";
    Score_limit=0.0;
    double maxsize=static_cast<double>(std::max(MainSequence.size(),Target.size()));
    Backtrace_stream.str(std::string());

    double size_diff=abs(MainSequence.size()- Target.size())/maxsize;

    if (Minimal_score_backtrace > 0.0) Record_backtrace=true;

    //We only process those that have comparable sizes
    if (size_diff <= Maximal_size_difference)
      {

//       NucleotideSequence<typename alignment_type::sequence_type> *changesequence=NULL;
//       std::string change;
//       if (MainSequence.size() >= Target.size())
//         {
//         prepare_main_sequence();
//         changesequence = &Target;
//         }
//       else
//         {
//         prepare_target_sequence();
//         changesequence = &MainSequence;
//         }

        
      if (reg_match(Directions,"[mrcd]"))
        {
        prepare_main_sequence();  
        
        if (reg_match(Directions,"m")) 
          {
          alignment_type::TopComment = " (genome 5'->3')"; alignment_type::BottomComment = " (primer 5'->3')";
          run_strand_alignment(&Target,"dir=m");
          TfRegLines.push_back(print_tfreg_dat("dir=m genome=5'->3' primer=5'->3'"));
          }
        Relative_scores[0]=relative_score();
      
        Target.reverse();
        if (reg_match(Directions,"r")) 
          {
          alignment_type::TopComment = " (genome 5'->3')"; alignment_type::BottomComment = " (primer 3'->5')";
          run_strand_alignment(&Target,"dir=r");
          TfRegLines.push_back(print_tfreg_dat("dir=r genome=5'->3' primer=3'->5'"));
          }
        Relative_scores[1]=relative_score();

        Target.complementary();
        if (reg_match(Directions,"d"))
          {
          alignment_type::TopComment = " (genome 5'->3')"; alignment_type::BottomComment = " (complementary primer 5'->3')";
          run_strand_alignment(&Target,"dir=d");
          TfRegLines.push_back(print_tfreg_dat("dir=d genome=5'->3' complementary_primer=5'->3'"));
          }
        Relative_scores[3]=relative_score();

        Target.reverse();
        if (reg_match(Directions,"c")) 
          {
          alignment_type::TopComment = " (genome 5'->3')"; alignment_type::BottomComment = " (complementary primer 3'->5')";
          run_strand_alignment(&Target,"dir=c");
          TfRegLines.push_back(print_tfreg_dat("dir=c genome=5'->3' complementary_primer 3'->5'"));
          }
        Relative_scores[2]=relative_score();

        //Changes back to original
        Target.complementary();
        }
	
      if (reg_match(Directions,"[MRCD]"))
        {
        prepare_main_sequence();  //to allocate space
        Target.reverse(); //3'-5'
        prepare_target_sequence();
        if (reg_match(Directions,"M")) 
          {
          alignment_type::TopComment = " (genome 5'->3')"; alignment_type::BottomComment = " (primer 3'->5')";
          run_strand_alignment(&MainSequence,"dir=M");
          TfRegLines.push_back(print_tfreg_dat("dir=M genome=5'->3' primer=3'->5'"));
          }

        MainSequence.reverse();
        if (reg_match(Directions,"R")) 
          {
          alignment_type::TopComment = " (genome 3'->5')"; alignment_type::BottomComment = " (primer 3'->5')";
          run_strand_alignment(&MainSequence,"dir=R");
          TfRegLines.push_back(print_tfreg_dat("dir=R genome=3'->5' primer=3'->5'"));
          }

        MainSequence.complementary();
        if (reg_match(Directions,"D")) 
          {
          alignment_type::TopComment = " (complementary genome 5'->3')"; alignment_type::BottomComment = " (primer 3'->5')";
          run_strand_alignment(&MainSequence,"dir=D");
          TfRegLines.push_back(print_tfreg_dat("dir=D complementary_genome=5'->3' primer=3'->5'"));
          }
          
        MainSequence.reverse();
        if (reg_match(Directions,"C")) 
          {
          alignment_type::TopComment = " (complementary genome 3'->5')"; alignment_type::BottomComment = " (primer 3'->5')";
          run_strand_alignment(&MainSequence,"dir=C");
          TfRegLines.push_back(print_tfreg_dat("dir=C complementary_genome 3'->5' primer=3'->5'"));
          }

        //Changes back to original
        MainSequence.complementary();
        }

      if (Minimal_report) print_minimal_report();
      else print_multi_strand_report();
      }
    else
      {
      if (ReportFile!=NULL and !Minimal_report)
        {
        *ReportFile << Target_description.substr(0,Target_description.find(' ')) << " "
                    << MainSequence_description.substr(0,MainSequence_description.find(' ')) << " "
                    << " SKIPPED ";
        if (Fasta_comment_type == "dmeltransposon")
          {
          *ReportFile << extract_field("name",MainSequence_description) << " "
                      << extract_field("name",Target_description) << " ";
          }
        *ReportFile << std::setw(5) << Target.size() << " "
                   << std::setw(5) << MainSequence.size() << std::endl;
        }
      }

    }

  /// \brief Tries to find best alignment over all strand directions
  void run_53_35_alignment(void)
    {
    Best_relative_score=-std::numeric_limits<typename alignment_type::score_type>::max();
    Best_strand="X ";
    Score_limit=0.0;
    double maxsize=static_cast<double>(std::max(MainSequence.size(),Target.size()));
    Backtrace_stream.str(std::string());

    double size_diff=abs(MainSequence.size()- Target.size())/maxsize;

    if (Minimal_score_backtrace > 0.0) Record_backtrace=true;

    //We only process those that have comparable sizes
    if (size_diff <= Maximal_size_difference)
      {

      NucleotideSequence<typename alignment_type::sequence_type> *changesequence=NULL;
      prepare_main_sequence();
      changesequence = &Target;

      
      changesequence->reverse();

      run_strand_alignment(changesequence,"MR");
      Relative_scores[1]=relative_score();
      Relative_scores[0]=Relative_scores[2]=Relative_scores[3]=0.0;


      //Changes back to original
      changesequence->reverse();
	
      if (Minimal_report) print_minimal_report();
      else print_multi_strand_report();
      }
    else
      {
      if (ReportFile!=NULL and !Minimal_report)
        {
        *ReportFile << Target_description.substr(0,Target_description.find(' ')) << " "
                    << MainSequence_description.substr(0,MainSequence_description.find(' ')) << " "
                    << " SKIPPED ";
        if (Fasta_comment_type == "dmeltransposon")
          {
          *ReportFile << extract_field("name",MainSequence_description) << " "
                      << extract_field("name",Target_description) << " ";
          }
        *ReportFile << std::setw(5) << Target.size() << " "
                   << std::setw(5) << MainSequence.size() << std::endl;
        }
      }

    }
  };//ends class


};//ends namespace
#endif

