// $Id: ContigTable.cpp 967 2017-07-13 12:04:25Z ge $
/// \file ContigTable.cpp
/// \brief The function bodies of class ContigTable
///
/// $Revision: 1.9 $
/// \author Gerald Weber <gweberbh@gmail.com>
#define CONTIGTABLE_CPP "$Id: ContigTable.cpp 967 2017-07-13 12:04:25Z ge $"

#include <algorithm>
#include <iterator>
#include "ContigTable.h"
using namespace std;
using namespace gbc;
namespace gbc
{


ContigTable::ContigTable(void)
  {
  Current=end();
  clear();
  Ignore_outside=true;
  Keyword_statistics.clear();
  }

void ContigTable::collect_keyword(const string &kw)
  {
  keyword_map::iterator loc=Keyword_statistics.find(kw);
  if (loc != Keyword_statistics.end()) loc->second++;
  else Keyword_statistics.insert(keyword_map::value_type(kw,1));
  }

void ContigTable::new_contig(ContigDescription &cd)
  {
  push_back(cd);
  collect_keyword(cd.Contig_type);
  try
    {
    Current=end(); Current--; // This is a bit risky, but cuts the processing time by half
    if (Current==end()) throw(__LINE__);
    }
  catch(int line)
    {
    cerr << endl<< "iterator Current out of range in " << __FILE__ << " line " << line << endl;
    exit(1);
    }
  }

void ContigTable::new_contig(const string &tp, const string &st)
  {
  ContigDescription cd;
  cd.description(st);
  cd.Contig_type=tp;
  new_contig(cd);
  }

bool ContigTable::find_contig(const string &tp, const string &st)
  {
  ContigTable::iterator ct=find(tp,st);
  bool found= ct!=end();
  if (found) Current=ct;
  return (found);
  }

void ContigTable::description(const string st)
  {
  try
    {
    if (Current==end()) throw;
    else current()->description(st);
    }
  catch(...)
    {
    cerr << "iterator Current out of range" << endl;
    exit(1);
    }
  }

string ContigTable::description(void)
  {
  try
    {
    if (Current==end()) throw;
    else return current()->description();
    }
  catch(...)
    {
    cerr << "iterator Current out of range" << endl;
    exit(1);
    }
  return string();
  }


bool ContigTable::in_current_contig_section(const streampos &ps)
  {
  if (Current != end())
    {
    return (current()->Stream_section.is_inside(ps));
    }
  else {return false;}
  }

bool ContigTable::in_current_contig_section(const nucleotide_position_type &ps)
  {
  if (Current != end())
    {
    return (current()->Contig_section.is_inside(ps));
    }
  else {return false;}
  }

bool ContigTable::between_nearest_contigs(const ContigTable::iterator &ct,const nucleotide_position_type &ps)
  {
  ContigTable::iterator next=ct; next++;
  if (next!=end())
    {return (ps > ct->Contig_section.end()) && (ps < next->Contig_section.begin());}
  else return false;
  }

bool ContigTable::inside_contig(const nucleotide_position_type &ns)
  {
  bool result=false;
  if (Current == end()) Current=begin();
  if (ns < (front().Contig_section.begin())) {/*Current=begin();*/ return false;}
  if (ns > (back().Contig_section.end()))    {Current=end(); Current--; return false;}
  if (in_current_contig_section(ns)) {return true;}
    else
    {
    ContigTable::iterator ct;
    for(ct=Current; ct != end(); ct++)
      {
      if (between_nearest_contigs(ct,ns)) {Current=ct; return false;}
        else if (ct->Contig_section.is_inside(ns)) {Current=ct; return true;}
      }
    for(ct=begin(); ct != Current; ct++)
      {
      if (between_nearest_contigs(ct,ns)) {Current=ct;  return false;}
        else if (ct->Contig_section.is_inside(ns)) {Current=ct; return true;}
      }
    }
  return result;
}

ContigTable::iterator ContigTable::mark(const bool consider, const std::string &tp, const std::string &st)
  {
  ContigTable::iterator ct,rt;
  bool any_tp= tp==std::string(), any_st= st==std::string();
  for (ct=begin(); ct!=end(); ct++) 
    {
    if (any_tp || (ct->Contig_type==tp)) 
      {
      if (any_st || (ct->description()==st)) 
        {ct->Ignore=!consider; rt=ct;}
      }
    }
  return rt;
  }

      bool ContigTable::ignore(const streampos &ps)
        {
        ContigTable::iterator ct;
        bool result=Ignore_outside;
        if (in_current_contig_section(ps))
          {
          result=Current->Ignore;
          }
        else
          {
          for(ct=begin(); ct != end(); ct++)
            {
            if ( (ps >= ct->Stream_section.begin()) && (ps <= ct->Stream_section.end()))
               {result=ct->Ignore; Current=ct; break;}
            }
          }
        return result;
	}

      bool ContigTable::ignore(const nucleotide_position_type &ps)
        {
	bool result=Ignore_outside;
 	if ((Contig_inside=inside_contig(ps))) result=current()->Ignore;
	Contig_begin=current()->Contig_section.begin()==ps;
	return result;
	}

     // This function is used in delete_ignored_contigs with the algorithm remove_if
     bool ignore_cd(ContigDescription &cd) {return cd.Ignore;}

     void ContigTable::delete_ignored_contigs(void)
        {
	try {remove_if(ignore_cd); Current=begin();}
	catch (...) {cerr << "Error in " << __FILE__ << " " << __LINE__ << endl;}
        }

      int ContigTable::count_overlaps(void)
        {
	int overlaps=0;
        ContigTable::iterator ct;
        for (ct=begin(); ct != end(); ct++)
          {
	  nucleotide_position_type ct_begin=ct->Contig_section.begin();
	  nucleotide_position_type ct_end  =ct->Contig_section.end();
	  ContigTable::iterator ct2=ct;
	  for (ct2++; ct2 != end(); ct2++)
	    {
	    nucleotide_position_type ct2_begin=ct2->Contig_section.begin();
	    nucleotide_position_type ct2_end  =ct2->Contig_section.end();
            if ( ((ct_begin > ct2_begin) && (ct_begin < ct2_end)) ||
	         ((ct_end   > ct2_begin) && (ct_end   < ct2_end)) )
		{
		overlaps++;
		//*(ct->Description->front())+=" overlaps";
		}
	    }
          }
        return overlaps;
	}


      int ContigTable::merge_overlaps(void)
        {
	int overlaps=0;
        ContigTable::iterator ct;
        for (ct=begin(); ct != end(); ct++)
          {
          if (!ct->Ignore)
	    {
	    nucleotide_position_type ct_begin=ct->Contig_section.begin();
	    nucleotide_position_type ct_end  =ct->Contig_section.end();
	    ContigTable::iterator ct2=ct; ct2++;
            while (ct2 != end())
	      {
	      if (!ct2->Ignore && (ct->Contig_type==ct2->Contig_type))
	        {
	        nucleotide_position_type ct2_begin=ct2->Contig_section.begin();
	        nucleotide_position_type ct2_end  =ct2->Contig_section.end();
                if ( ((ct_begin > ct2_begin) && (ct_begin < ct2_end)) ||
	             ((ct_end   > ct2_begin) && (ct_end   < ct2_end)) )
		  {
                  if (ct_begin > ct2_begin) ct->Contig_section.begin(ct2_begin);
		  if (ct_end   < ct2_end)   ct->Contig_section.end(ct2_end);
		  ct->description(ct->description()+string(" | ")+ct2->description());
		  ct2=erase(ct2);
		  overlaps++;
		  }
		else ct2++;
		}
              else ct2++;
              }
	    }
          }
	return overlaps;
	}

      std::streampos ContigTable::next_streampos(const std::streampos &ps, bool forward)
        {
	std::streampos newpos=ps;
	if (!ignore(ps)) newpos=ps;
	else
	  {
	  ContigTable::iterator ct;
	  if (forward)
	    {
	    for(ct=begin(); ct != end(); ct++)
	      {
	      if ( (ct->Stream_section.begin() >= ps) && (!ct->Ignore))
	        {
                newpos=ct->Stream_section.begin();
                Current=ct;
	        break;
		}
	      }
	    }
          else
	    {
	    ContigTable::reverse_iterator ct;
	    for(ct=rbegin(); ct != rend(); ct++)
	      {
	      if ( (ct->Stream_section.end() <= ps) && (!ct->Ignore))
	        {
                newpos=ct->Stream_section.end();
                Current=ct.base(); Current--;
	        break;
		}
	      }
	    if (ct==rend()) newpos=(streampos)(-1);
	    }
	  }
	return newpos;
	}

      nucleotide_position_type ContigTable::next_contigpos(const nucleotide_position_type &ps)
        {
	nucleotide_position_type newpos=ps;
	if (!ignore(ps)) newpos=ps;
	else
	  {
	  ContigTable::iterator ct;
	  for(ct=begin(); ct != end(); ct++)
	    {
	    if ( (ct->Contig_section.begin() >= ps) && (!ct->Ignore))
	      {
              newpos=ct->Contig_section.begin();
              Current=ct;
	      break;
	      }
	    }
	  }
	return newpos;
	}

std::ostream& operator<<(std::ostream &out, const ContigTable &ct)
  {
  std::copy(ct.begin(),ct.end(),std::ostream_iterator<ContigDescription>(out,"\n"));
  return out;
  }

std::ostream& operator<<(std::ostream &out, const ContigTable::keyword_map &km)
  {
  ContigTable::keyword_map::const_iterator seq;
  for(seq=km.begin(); seq != km.end(); seq++) out << seq->first << " " << seq->second << endl;
  return out;
  }

};

